﻿EnableExplicit

Structure _record
  is_new.i
  timestamp.q
  bytebuffer.a[10]
  mode.a
  exponent.a
  value.w
  cs.i
  min.w
  max.w
  dvalue.d
  divider.w
  precision.a
  unit.s
EndStructure

Enumeration STATES
  #WAIT_FIRST_START_BYTE
  #WAIT_SECOND_START_BYTE
  #WAIT_THIRD_START_BYTE
  #WAIT_MODE_BYTE
  #WAIT_EXPONENT_BYTE
  #WAIT_HIGH_VALUE_BYTE
  #WAIT_MIDDLE_VALUE_BYTE
  #WAIT_LOW_VALUE_BYTE
  #WAIT_HIGH_CS_BYTE
  #WAIT_LOW_CS_BYTE
  #MAKE_DATA_ANALYSIS
EndEnumeration

Enumeration MODE
  #SQUAREWAVE         = $03
  #VOLT_AC            = $15
  #VOLT_DC            = $16
  #AMPER              = $18
  #AMPER_MILLI        = $19
  #AMPER_MICRO        = $1A
  #CONTINUITY_OHM     = $1B
  #DIODE_VOLT         = $1C
  #RESISTANCE_OHM     = $1D
EndEnumeration

Enumeration RANGE
  #RANGE_NONE         = $00
  #RANGE_A            = $01
  #RANGE_B            = $02
  #RANGE_C            = $04
  #RANGE_D            = $08
  #RANGE_E            = $10
  #RANGE_F            = $20
  #RANGE_G            = $40
  #RANGE_H            = $80
EndEnumeration

Enumeration PORT_STATUS
  #CLOSE
  #OPEN
EndEnumeration

Enumeration SAVE_STATUS
  #SAVE
  #NO_SAVE
EndEnumeration

Enumeration FORM_WINDOW
  #WIN_MAIN
EndEnumeration

Enumeration FORM_GADGET
  #IMG_VALUE
  #PGB_VALUE
  #FRM_DEVICE
  #TXT_PORT
  #STR_PORT
  #BTN_PORT
  #FRM_SAVE
  #BTN_SAVE
  #TXT_INTERVAL
  #STR_INTERVAL
  #TXT_INFO_1
  #TXT_INFO_2
EndEnumeration

Enumeration FORM_FONT
  #FONT_WIN_MAIN_0
  #FONT_WIN_MAIN_1
EndEnumeration

Global record._record
Global strCom.s               = "COM1"
Global hCom.i                 = #Null
Global receive_thread_sync.i  = CreateMutex()
Global receive_thread_abort.i = #False
Global h_receive_thread.i     = #Null
Global time.q                 = #Null
Global time_thread_abort.i    = #True
Global h_time_thread.i        = #Null
Global hFile.i                = #Null
Global status_port.i          = #CLOSE
Global status_save.i          = #NO_SAVE
Global status_work.i          = #False
Global recvrecords.i          = #Null
Global saverecords.i          = #Null
Global save_interval.i

LoadFont(#FONT_WIN_MAIN_0,"Arial", 14)
LoadFont(#FONT_WIN_MAIN_1,"Arial", 10)

Procedure openwin_main(x.i = 0, y.i = 0, width.i = 467, height.i = 467)
  OpenPreferences(GetPathPart(ProgramFilename()) + GetFilePart(ProgramFilename(), #PB_FileSystem_NoExtension) + ".ini")
  x             = ReadPreferenceInteger("WIN_X", 0)
  y             = ReadPreferenceInteger("WIN_Y", 0)  
  strCom        = ReadPreferenceString ("COM", "COM3")
  save_interval = ReadPreferenceInteger("INTERVAL", 4)
  ClosePreferences()
  
  OpenWindow(#WIN_MAIN, x, y, width, height, "Multimeter", #PB_Window_SystemMenu | #PB_Window_MinimizeGadget)
  CanvasGadget(#IMG_VALUE, 7, 7, 453, 233)
  ProgressBarGadget(#PGB_VALUE, 7, 247, 453, 17, 0, 0)
  SetGadgetColor(#PGB_VALUE, #PB_Gadget_FrontColor,RGB(0,128,0))
  FrameGadget(#FRM_DEVICE, 7, 273, 453, 80, " Schnittstelle ")
  SetGadgetFont(#FRM_DEVICE, FontID(#Font_win_main_0))
  TextGadget(#TXT_PORT, 20, 313, 53, 27, "Port:")
  SetGadgetFont(#TXT_PORT, FontID(#Font_win_main_0))
  StringGadget(#STR_PORT, 73, 307, 293, 33, strCom)
  SetGadgetFont(#STR_PORT, FontID(#Font_win_main_0))
  ButtonGadget(#BTN_PORT, 380, 307, 67, 33, "Start")
  SetGadgetFont(#BTN_PORT, FontID(#Font_win_main_0))
  FrameGadget(#FRM_SAVE, 7, 358, 453, 75, " Datenaufzeichnung ")
  SetGadgetFont(#FRM_SAVE, FontID(#Font_win_main_0))
  ButtonGadget(#BTN_SAVE, 287, 387, 160, 33, "Start")
  SetGadgetFont(#BTN_SAVE, FontID(#Font_win_main_0))
  TextGadget(#TXT_INTERVAL, 20, 393, 153, 33, "Intervall: 500ms *")
  SetGadgetFont(#TXT_INTERVAL, FontID(#Font_win_main_0))
  StringGadget(#STR_INTERVAL, 173, 387, 100, 33, Str(save_interval), #PB_String_Numeric)
  SetGadgetFont(#STR_INTERVAL, FontID(#Font_win_main_0))
  TextGadget(#TXT_INFO_1, 7, 440, 220, 20, "Empfangen: 0", #PB_Text_Border)
  SetGadgetFont(#TXT_INFO_1, FontID(#Font_win_main_1))
  TextGadget(#TXT_INFO_2, 240, 440, 220, 20, "Gespeichert: 0", #PB_Text_Border)
  SetGadgetFont(#TXT_INFO_2, FontID(#Font_win_main_1))
EndProcedure

Procedure.w range_min(mode.a, range.a)
  Protected retvalue.w = -1999
  If (mode = #VOLT_DC And range = #RANGE_F) Or (mode = #VOLT_AC And range = #RANGE_F)
    retvalue -300
  ElseIf mode = #AMPER And range = #RANGE_G
    retvalue -1000
  ElseIf mode = #RESISTANCE_OHM Or mode = #CONTINUITY_OHM Or mode = #DIODE_VOLT
    retvalue = 0
  EndIf
  ProcedureReturn retvalue
EndProcedure

Procedure.w range_max(mode.a, range.a)
  Protected retvalue.w = 1999
  If (mode = #Volt_DC And range = #RANGE_F) Or (mode = #Volt_AC And range = #RANGE_F)
    retvalue = 300
  ElseIf mode = #AMPER And range = #RANGE_G
    retvalue = 1000
  EndIf
  ProcedureReturn retvalue
EndProcedure

Procedure.b is_overloaded(mode.a, value.w, min.w, max.w)
  Protected retvalue.b
  If mode = #SQUAREWAVE
    retvalue = #True
  ElseIf value < min Or value > max
    retvalue = #True
  Else
    retvalue = #False
  EndIf
  ProcedureReturn retvalue
EndProcedure

Procedure time_thread(dummy.i)
  time = Date() * 1000
  Protected interval.i = ElapsedMilliseconds()
  Protected last_interval.i = interval
  Repeat
    interval = ElapsedMilliseconds() - last_interval
    last_interval + interval
    time + interval
    Delay(9)
  Until time_thread_abort
EndProcedure

Procedure create_thread_time()
  If Not h_time_thread
    time_thread_abort = #False
    h_time_thread = CreateThread(@time_thread(), 42)
  EndIf
EndProcedure

Procedure stop_thread_time()
  If h_time_thread
    time_thread_abort = #True
    WaitThread(h_time_thread)
    h_time_thread = #Null
  EndIf
EndProcedure

Procedure receive_thread(dummy.i)
  If hCom
    Protected state.i = #WAIT_FIRST_START_BYTE
    Protected inbyte.a, mode.a, exponent.a, value.i, cs.i, bcs.i, a.i
    Protected Dim buffer.a(9)
    Repeat
      If AvailableSerialPortInput(hCom)
        If ReadSerialPortData(hCom, @inbyte, 1) <> 1
          state = #WAIT_FIRST_START_BYTE
        Else
          Debug Hex(inbyte)
          If state = #WAIT_FIRST_START_BYTE And inbyte = $DC
            buffer(0) = inbyte
            state = #WAIT_SECOND_START_BYTE
          ElseIf state = #WAIT_SECOND_START_BYTE And inbyte = $BA
            buffer(1) = inbyte
            state = #WAIT_THIRD_START_BYTE
          ElseIf state = #WAIT_THIRD_START_BYTE And inbyte = $01
            buffer(2) = inbyte
            cs = inbyte 
            state = #WAIT_MODE_BYTE
          ElseIf state = #WAIT_MODE_BYTE
            buffer(3) = inbyte
            mode = inbyte
            cs = (cs + inbyte) & $FFFF
            state = #WAIT_EXPONENT_BYTE
          ElseIf state = #WAIT_EXPONENT_BYTE
            buffer(4) = inbyte
            exponent = inbyte
            cs = (cs + inbyte) & $FFFF
            state = #WAIT_HIGH_VALUE_BYTE
          ElseIf state = #WAIT_HIGH_VALUE_BYTE
            buffer(5) = inbyte
            ;value = inbyte * $10000                                  ; eher nicht
            cs = (cs + inbyte) & $FFFF
            state = #WAIT_MIDDLE_VALUE_BYTE
          ElseIf state = #WAIT_MIDDLE_VALUE_BYTE
            buffer(6) = inbyte
            value = (inbyte * $100)
            cs = (cs + inbyte) & $FFFF
            state = #WAIT_LOW_VALUE_BYTE
          ElseIf state = #WAIT_LOW_VALUE_BYTE
            buffer(7) = inbyte
            value + inbyte
            cs = (cs + inbyte) & $FFFF
            state = #WAIT_HIGH_CS_BYTE
          ElseIf state = #WAIT_HIGH_CS_BYTE
            buffer(8) = inbyte
            bcs = inbyte * $100
            state = #WAIT_LOW_CS_BYTE
          ElseIf state = #WAIT_LOW_CS_BYTE
            buffer(9) = inbyte
            bcs + inbyte
            state = #MAKE_DATA_ANALYSIS
          Else
            state = #WAIT_FIRST_START_BYTE
          EndIf
        EndIf
        If state = #MAKE_DATA_ANALYSIS
          If bcs = cs
            LockMutex(receive_thread_sync)
            record\is_new = #True
            record\timestamp = time
            For a = 0 To 9 : record\bytebuffer[a] = buffer(a) : Next a
            record\mode = mode
            record\exponent = exponent
            record\value = value
            record\cs = cs
            record\min = range_min(mode, exponent)
            record\max = range_max(mode, exponent)
            If is_overloaded(mode, value, record\min, record\max)
              record\dvalue = 9999
              record\unit = ""
            ElseIf mode = #VOLT_DC
              If exponent = #RANGE_B
                record\precision = 1
                record\divider = 10
                record\unit = "mV DC"
              ElseIf exponent = #RANGE_C
                record\precision = 3
                record\divider = 1000
                record\unit = "V DC"
              ElseIf exponent = #RANGE_D
                record\precision = 2
                record\divider = 100
                record\unit = "V DC"
              ElseIf exponent = #RANGE_E
                record\precision = 1
                record\divider = 10
                record\unit = "V DC"
              ElseIf exponent = #RANGE_F
                record\precision = 0
                record\divider = 1
                record\unit = "V DC"
              EndIf
              record\dvalue = value / record\divider
            ElseIf mode = #VOLT_AC
              If exponent = #RANGE_C
                record\precision = 3
                record\divider = 1000
                record\unit = "V AC"
              ElseIf exponent = #RANGE_D
                record\precision = 2
                record\divider = 100
                record\unit = "V AC"
              ElseIf exponent = #RANGE_E
                record\precision = 1
                record\divider = 10
                record\unit = "V AC"
              ElseIf exponent = #RANGE_F
                record\precision = 0
                record\divider = 1
                record\unit = "V AC"
              EndIf
              record\dvalue = value / record\divider
            ElseIf mode = #AMPER
              If exponent = #RANGE_F
                record\precision = 3
                record\divider = 1000
                record\unit = "A"
              ElseIf exponent = #RANGE_G
                record\precision = 2
                record\divider = 100
                record\unit = "A"
              EndIf
              record\dvalue = value / record\divider
            ElseIf mode = #AMPER_MILLI
              If exponent = #RANGE_D
                record\precision = 2
                record\divider = 100
                record\unit = "mA"
              ElseIf exponent = #RANGE_E
                record\precision = 1
                record\divider = 10
                record\unit = "mA"
              ElseIf exponent = #RANGE_F
                record\precision = 3
                record\divider = 1000
                record\unit = "mA"
              EndIf
              record\dvalue = value / record\divider
            ElseIf mode = #AMPER_MICRO
              If exponent = #RANGE_A
                record\unit = "µA"  
              ElseIf exponent = #RANGE_B
                record\precision = 1
                record\divider = 10
                record\unit = "µA"
              ElseIf exponent = #RANGE_C
                record\precision = 0
                record\divider = 1
                record\unit = "µA"
              EndIf
              record\dvalue = value / record\divider
            ElseIf mode = #RESISTANCE_OHM
              If exponent = #RANGE_A
                record\precision = 1
                record\divider = 10
                record\unit = "Ohm"
              ElseIf exponent = #RANGE_B
                record\precision = 3
                record\divider = 1000
                record\unit = "kOhm"
              ElseIf exponent = #RANGE_C
                record\precision = 2
                record\divider = 100
                record\unit = "kOhm"
              ElseIf exponent = #RANGE_D
                record\precision = 1
                record\divider = 10
                record\unit = "kOhm"
              ElseIf exponent = #RANGE_E
                record\precision = 3
                record\divider = 1000
                record\unit = "MOhm"
              ElseIf exponent = #RANGE_F
                record\precision = 2
                record\divider = 100
                record\unit = "MOhm"
              EndIf
              record\dvalue = value / record\divider
            ElseIf mode = #CONTINUITY_OHM
              If exponent = #RANGE_A
                record\precision = 1
                record\divider = 10
                record\unit = "Ohm"
                record\dvalue = value / record\divider
              EndIf
            ElseIf mode = #DIODE_VOLT
              If exponent = #RANGE_C
                record\precision = 3
                record\divider = 1000
                record\unit = "V"
              EndIf
              record\dvalue = value / record\divider
            ElseIf mode = #SQUAREWAVE
              record\unit = ""
            EndIf
            UnlockMutex(receive_thread_sync)
          EndIf
          state = #WAIT_FIRST_START_BYTE
        EndIf
      Else
        Delay(50)
      EndIf
    Until receive_thread_abort <> #False
  EndIf
EndProcedure

Procedure show_value(dvalue.d, precision.a, unit.s, mode.a)
  Protected strvalue.s
  Static old_dvalue.d = -9999
  Static old_mode.a = $FF
  Static old_precision.a = $FF
  If old_dvalue <> dvalue Or old_mode <> mode Or old_precision <> precision
    If StartVectorDrawing(CanvasVectorOutput(#IMG_VALUE))
      VectorSourceColor(RGBA(255, 255, 255, 255))
      FillVectorOutput()
      VectorSourceColor(RGBA(0, 0, 0, 255))
      If mode = #SQUAREWAVE
        VectorFont(FontID(0), 50)
        strvalue = "SQUAREWAVE"
      Else
        VectorFont(FontID(0), 150)
        If dvalue < 2000
          strvalue = StrD(dvalue, precision)
        Else
          strvalue = "O.L"
        EndIf
      EndIf
      MovePathCursor(453/2 - VectorTextWidth(strvalue)/2, 0)
      DrawVectorText(strvalue)
      VectorFont(FontID(0), 75)
      MovePathCursor(453/2 - VectorTextWidth(unit)/2, 150)
      DrawVectorText(unit)
      StopVectorDrawing()
    EndIf
    old_dvalue = dvalue
    old_mode = mode
    old_precision = precision
  EndIf
EndProcedure

Procedure.s make_value_str(dvalue.d, precision.a, mode.a)
  Protected strvalue.s
  If mode = #SQUAREWAVE
    strvalue = "SQUAREWAVE"
  Else
    If dvalue < 2000
      strvalue = StrD(dvalue, precision)
    Else
      strvalue = "O.L"
    EndIf
  EndIf
  ProcedureReturn strvalue
EndProcedure

Procedure create_thread_receive()
  If Not h_receive_thread
    receive_thread_abort = #False
    h_receive_thread = CreateThread(@receive_thread(), 42)
    If IsThread(h_receive_thread)
      SetGadgetText(#BTN_PORT, "Stop")
      recvrecords = 0
      SetGadgetText(#TXT_INFO_1, "Empfangen: " + Str(recvrecords))
      status_port = #OPEN
    Else
      MessageRequester("Fehler", "Kann Thread nicht erstellen.", #PB_MessageRequester_Error)
      h_receive_thread = #Null
    EndIf
  Else
    MessageRequester("Fehler", "Kann Thread nicht erstellen.", #PB_MessageRequester_Error)  
  EndIf
EndProcedure

Procedure stop_thread_receive()
  If h_receive_thread
    receive_thread_abort = #True
    WaitThread(h_receive_thread)
    h_receive_thread = #Null
  EndIf  
EndProcedure

Procedure open_file()
  save_interval = Val(GetGadgetText(#STR_INTERVAL)) * 500
  If status_port = #OPEN
    Protected filename.s = SaveFileRequester("Speichern in Datei", "log.csv", "*.*", 0)
    If filename
      hFile = OpenFile(#PB_Any, filename, #PB_File_SharedRead | #PB_File_Append | #PB_UTF8)
      If hFile
        SetGadgetText(#BTN_SAVE, "Stop")
        status_save = #SAVE
        saverecords = 0
        SetGadgetText(#TXT_INFO_2, "Gespeichert: " + Str(saverecords))
      Else
        MessageRequester("Fehler", "Kann Datei " + GetFilePart(filename) + " nicht oeffnen.", #PB_MessageRequester_Error)
      EndIf
    EndIf
  Else
    MessageRequester("Fehler", "Kein Port geoeffnet.", #PB_MessageRequester_Error)
  EndIf
EndProcedure

Procedure close_file()
  If hFile
    CloseFile(hFile)
    hFile = #Null
    SetGadgetText(#BTN_SAVE, "Start")
    status_save = #NO_SAVE
  EndIf
EndProcedure

Procedure open_serial(strCom.s)
  hCom = OpenSerialPort(#PB_Any, strCom, 38600, #PB_SerialPort_NoParity, 8, 1, #PB_SerialPort_NoHandshake, 1024, 1024)
  If hCom
    create_thread_receive()
  Else
    MessageRequester("Fehler", "Kann Schnittstelle " + strCom + " nicht oeffnen.", #PB_MessageRequester_Error)
  EndIf
EndProcedure

Procedure close_serial()
  If hCom
    CloseSerialPort(hCom)
    hCom = #Null
  EndIf
EndProcedure

Procedure.q get_timestamp()
  ProcedureReturn time
EndProcedure

Procedure.s make_timestamp_str(timestamp.q)
  ProcedureReturn FormatDate("%yyyy.%mm.%dd %hh:%ii:%ss", timestamp / 1000) + "." + RSet(Str(timestamp % 1000), 3, "0")
EndProcedure

Procedure main()
  Protected win_main_quit.i  = #False
  Protected ausgabetxt.s     = #Null$
  Protected winEvent.i, gadEvent.i, a.i, next_save.q = 0, timestamp.q
  openwin_main()
  create_thread_time()
  Repeat
    winEvent = WindowEvent()
    If winEvent = #PB_Event_CloseWindow
      status_work = #True
      stop_thread_receive()
      stop_thread_time()
      close_file()
      close_serial()
      win_main_quit = #True
    ElseIf winEvent =  #PB_Event_Gadget
        status_work = #True
        gadEvent = EventGadget()
        Select gadEvent
          Case #BTN_PORT                                                                                              
            If status_port = #OPEN                                                                      ; close port
              stop_thread_receive()
              close_file()
              close_serial()
              SetGadgetText(#BTN_PORT, "Start")
              status_port = #CLOSE
            Else                                                                                        ; open port
              strCom = GetGadgetText(#STR_PORT)
              open_serial(strCom)
            EndIf
          Case #BTN_SAVE
            If status_save = #NO_SAVE                                                                   ; open file
              open_file()
            Else                                                                                        ; close file
              close_file()
            EndIf
        EndSelect
    EndIf
    If TryLockMutex(receive_thread_sync)
      If record\is_new
        status_work = #True
        show_value(record\dvalue, record\precision, record\unit, record\mode)
        recvrecords + 1
        SetGadgetText(#TXT_INFO_1, "Empfangen: " + Str(recvrecords))
        record\is_new = #False
      EndIf
      If hFile
        If next_save < time
          next_save = time + save_interval
          timestamp.q = get_timestamp()
          ausgabetxt = Str(timestamp) + ";" + make_timestamp_str(timestamp) + ";"
          For a = 0 To 9 : ausgabetxt + RSet(Hex(record\bytebuffer[a]), 2, "0") : Next a
          ausgabetxt + ";" + make_value_str(record\dvalue, record\precision, record\mode) + ";" + record\unit
          WriteStringN(hFile, ausgabetxt, #PB_UTF8)
          saverecords + 1
          SetGadgetText(#TXT_INFO_2, "Gespeichert: " + Str(saverecords))
          status_work = #True
        EndIf
      EndIf
      UnlockMutex(receive_thread_sync)
    EndIf
    If status_work = #True
      status_work = #False
    Else
      Delay(10)
      status_work = #False
    EndIf
  Until win_main_quit
  If CreatePreferences(GetPathPart(ProgramFilename()) + GetFilePart(ProgramFilename(), #PB_FileSystem_NoExtension) + ".ini")
    WritePreferenceInteger("WIN_X", WindowX(#WIN_MAIN))
    WritePreferenceInteger("WIN_Y", WindowY(#WIN_MAIN))
    WritePreferenceString ("COM", GetGadgetText(#STR_PORT))
    WritePreferenceInteger("INTERVAL", Val(GetGadgetText(#STR_INTERVAL)))
    ClosePreferences()
  EndIf
EndProcedure

main()
End


  



; IDE Options = PureBasic 5.72 (Windows - x64)
; CursorPosition = 518
; FirstLine = 483
; Folding = ----
; EnableThread
; Executable = ..\..\Desktop\Multimeter.exe